!***********************************************************************
! Module: ommosaic_mod
!
! Authors:  Christopher Brindle - SSC (MSSL)
!          
! Description: Mosaics OM images
! 
! Comments: Based on emosiac_mod.f90 written by R. West - SSC (Leicester)
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!************************************************************************ 
module ommosaic_mod

  use dal
  use param  
  use errorHandling
  use constants
  use sla

  implicit none 
  
  private 

  type ImageT
     character (len=50) :: filename
     real (kind=single), dimension(:, :), pointer :: image
     real (kind=single), dimension(:, :), pointer :: weight
  
     real (kind=double) :: xrpix, yrpix, xrval, yrval
     real (kind=double) :: xdelt, ydelt, roll
     integer :: datatype, dims(0 : 1), xmin, xmax, ymin, ymax
     real (kind=double) :: sky2img(3, 3)
  end type ImageT

  public ommosaic
   
contains 
  
!************************************************************
subroutine ommosaic
!************************************************************
    type (ImageT), allocatable, dimension(:) :: input
    type (ImageT) :: output
   
    integer i, noFiles
    character(len = 100), allocatable, dimension(:) :: fileName
    character (len = 100) :: outputFile
    integer :: maxIndex
    real(kind=single) :: maxPixelSize   
    noFiles = parameterCount("imagesets")
    allocate(fileName(1 : noFiles))
    
    do i = 1, noFiles
        fileName(i) = stringParameter("imagesets", i - 1)
    end do 

    outputFile = stringParameter("mosaicedset")
 
       

    !****************************************************       
    ! Open input files, read data arrays and WCS keywords
    !****************************************************
    

    allocate(input(1 : noFiles))
    maxPixelSize = 0.0
    maxIndex = 1
    do i = 1, noFiles
        call readImage(input(i), fileName(i))
        !****************************************
        ! Check the pixel size
        !***************************************
        if (input(i)%xdelt .gt. maxPixelSize) then
            maxIndex = i
            maxPixelSize = input(i)%xdelt
        end if    
    end do           

    !******************************************
    ! Chacek that the filters are all the same
    !******************************************
    call CheckFilters(noFiles, fileName)
       
    !**************************************************************
    ! For now assume output image has same reference coordinate and
    ! pixel size as the first input image
    !*************************************************************
    output%xrval = input(maxIndex)%xrval
    output%xdelt = input(maxIndex)%xdelt
    output%yrval = input(maxIndex)%yrval
    output%ydelt = input(maxIndex)%ydelt
    output%roll  = input(maxIndex)%roll
       
    !*******************************************************
    ! Calculate size of output image (physical and in pixels)
    !*******************************************************
    call calculateOutputSize(noFiles, input, output)

                                                                                                                        
    ! Loop through input images adding to the output
    do i = 1, noFiles
        call addImage(input(i), output)
    end do
    deallocate(input)       
 
    !*****************************************************
    ! Normalise each pixel image by the sum of the 
    ! exposure times for each pixel
    !*****************************************************
    where (output%weight > 0.0) output%image = output%image / output%weight

    call writeImage(output, outputFile)        

    !******************
    ! Release memory
    !******************
   

    deallocate(fileName)
   
end subroutine ommosaic

!*******************************************************************
! Calculate size of output image in physical units and in
! pixels. Achieves this by casting the coordinates of the corner
! pixels into the frames of the first image.
!*******************************************************************
subroutine calculateOutputSize(nInput, input, output)
    integer, intent(in) :: nInput
    type (ImageT), intent(inout) :: input(nInput), output
   

    ! Local variables
    integer :: i, j, k, xmin, xmax, ymin, ymax
    real (kind=double) :: az, el, axvector(3), vlocal(3), vsky(3),&
         & xpix, ypix, outaz, outel

    ! Initialise bounds (in pixels)
    xmin = 0; 
    xmax = 0; 
    ymin = 0; 
    ymax = 0

    ! Create local -> celestial transformation matrix for output image
    axvector(1) = output%xrval * DegToRad
    axvector(2) = output%yrval * DegToRad
    axvector(3)= output%roll * DegToRad
    call sla_deuler('ZYX', axvector(1), -axvector(2), axvector(3),&
         & output%sky2img)

    !****************************************************************
    ! Loop through the input images, casting each of the four corners
    ! into the frame of the output image. This allows us to find the
    ! bounding box of the input images in the output frame.
    !****************************************************************
    do i = 1, nInput
       axvector(1) = input(i)%xrval * DegToRad
       axvector(2) = input(i)%yrval * DegToRad
       axvector(3) = input(i)%roll * DegToRad
       call sla_deuler('ZYX', axvector(1), -axvector(2), axvector(3),&
            & input(i)%sky2img)

       ! Initialise bounding box
       input(i)%xmin = 0
       input(i)%xmax = 0
       input(i)%ymin = 0
       input(i)%ymax = 0

       ! NB. There's a half-pixel shift in here somewhere which needs to
       ! be sorted out (thanks FITS people. No, really, thanks).
       !
       
       ! Loop through each corner
       do j = 0, 1
          az = (input(i)%dims(0) * j - input(i)% xrpix + 0.5) * input(i)%xdelt

          do k = 0, 1
             el = (input(i)%dims(1) * k - input(i)%yrpix + 0.5) * input(i)%ydelt

             !******************************************************* 
             ! Transform corner from local frame to output frame. I'm
             ! doing this the long-handed (and stupid) way as it's
             ! easier to get ones head around.
             !******************************************************

             ! Convert spherical -> cartesian
             call sla_dcs2c(az * DegToRad, el * DegToRad, vlocal)

             ! Transform from local to celestial (cartesian)
             call sla_dimxv(input(i)%sky2img, vlocal, vsky)
             
             ! Transform into output frame
             call sla_dmxv(output%sky2img, vsky, vlocal)

             ! Convert to back spherical coordinates
             call sla_dcc2s(vlocal, outaz, outel)

             ! Convert into pixels in the output frame
             xpix = outaz * RadToDeg / output%xdelt
             ypix = outel * RadToDeg / output%ydelt

             ! Check against global bounds
             xmin = min(xmin, floor(xpix))
             xmax = max(xmax, ceiling(xpix))
             ymin = min(ymin, floor(ypix))
             ymax = max(ymax, ceiling(ypix))

             ! Update local bounds
             input(i)%xmin = min(input(i)%xmin, floor(xpix))
             input(i)%xmax = max(input(i)%xmax, ceiling(xpix))
             input(i)%ymin = min(input(i)%ymin, floor(ypix))
             input(i)%ymax = max(input(i)%ymax, ceiling(ypix))
          end do
       end do
    end do

    ! Update reference pixel position of output image
    output%xrpix = -xmin
    output%yrpix = -ymin
    

    ! Calculate size of image, allocate array
    output%dims(0) = (xmax - xmin + 1)
    output%dims(1) = (ymax - ymin + 1)

    allocate(output%image(output%dims(0), output%dims(1)))
    output%image = 0.0

    allocate(output%weight(output%dims(0), output%dims(1)))
    output%weight = 0.0

    ! Translate bounds of all input images
    do i = 1, nInput
       input(i)%xmin = input(i)%xmin - xmin
       input(i)%xmax = input(i)%xmax - xmin
       input(i)%ymin = input(i)%ymin - ymin
       input(i)%ymax = input(i)%ymax - ymin
    end do
     
end subroutine calculateOutputSize
!********************************************************
! Read image and associated WCS keywords from a dataset
!
!*********************************************************
subroutine readImage(image, imagefile)
    type (ImageT), intent(out) :: image
    character (len=*), intent(in) :: imagefile
    

    ! Local variables
    integer, dimension(:), pointer :: dims
    type (DataSetT) :: set
    type (ArrayT) :: arr

    ! Open dataset, find image array
    image%filename = imagefile

    set = dataset(imagefile, read)
    arr = array(set, 'PRIMARY', read)

    ! Check dimensions of image
    if(numberOfDimensions(arr) /= 2)&
         & call error('badDims', 'Only 2-D images are supported')
    dims=>dimensions(arr)
    allocate(image%image(dims(0), dims(1)))
    allocate(image%weight(dims(0), dims(1)))
    
    
    image%dims = dims

    ! Read WCS keywords
    image%xrpix = real64Attribute(arr, 'CRPIX1')
    image%xrval = real64Attribute(arr, 'CRVAL1')
    image%xdelt = real64Attribute(arr, 'CDELT1')
    image%yrpix = real64Attribute(arr, 'CRPIX2')
    image%yrval = real64Attribute(arr, 'CRVAL2')
    image%ydelt = real64Attribute(arr, 'CDELT2')
    image%roll = 0.0
    
    

    ! Perform type conversion on image data
    image%datatype = datatype(arr)
    call readImageAsReal32(arr, image%image)


    !*******************************
    ! Read in the exposure image
    !*******************************
    arr = array(set, 'EXPOSURE', read)
    call readImageAsReal32(arr, image%weight)


    !******************************
    ! Read the exposure map
    !******************************

    
    ! Free the dataset
    call release(arr)
    call release(set)

  end subroutine readImage

!********************************************
! Add an image to the output
!
!********************************************
subroutine addImage(input, output)
    type (ImageT), intent(in) :: input
    type (ImageT), intent(inout) :: output
    
    
    ! Local variables
    integer :: i, j, xpix, ypix
    real (kind=double) :: iaz, iel, oaz, oel, vlocal(3), vsky(3)
   
    !****************************************************************
    ! Loop over every pixel in the output image, transform from local
    ! coordinates to sky coordinates, transform from sky coordinates
    ! local coordinates in the input image, pick up pixel value and
    ! copy to output pixel
    !
    ! The matrix stuff here could be optimised quite a bit I think
    !***************************************************************
    do j = 1, output%dims(1)
       oel = (j - output%yrpix) * output%ydelt

       do i = 1, output%dims(0)
          oaz = (i - output%xrpix) * output%xdelt

          ! Convert spherical -> cartesian
          call sla_dcs2c(oaz * DegToRad, oel * DegToRad, vlocal)

          ! Transform from local to celestial (cartesian)
          call sla_dimxv(output%sky2img, vlocal, vsky)

          ! Transform into output frame
          call sla_dmxv(input%sky2img, vsky, vlocal)

          ! Convert to back spherical coordinates
          call sla_dcc2s(vlocal, iaz, iel)

          ! Convert into pixels in the output frame
          xpix = nint(iaz * RadToDeg / input%xdelt + input%xrpix)
          ypix = nint(iel * RadToDeg / input%ydelt + input%yrpix)

          ! Check against bounds of input image
          if(xpix > 0 .and. xpix < input%dims(0) .and. &
               & ypix > 0 .and. ypix < input%dims(1)) then

              output%image(i, j) = output%image(i, j) + &
                     & input%image(xpix, ypix)

              !************************************************
              ! Sum up the exposure times for the normalisation
              !************************************************
              output%weight(i, j) = output%weight(i, j) + input%weight(xpix, ypix)
          end if

       end do
    end do

end subroutine addImage
!**************************************************
! Write image to a new file
!**************************************************
subroutine writeImage(image, filename)
    type (imageT), intent(in) :: image
    character (len=*), intent(in) :: filename

    ! Local variables
    type (DataSetT) :: set
    type (ArrayT) :: arr
    real (kind=single), dimension(:,:), pointer :: optr

    ! Create new dataset
    set = dataSet(filename, CREATE)

    ! Add primary array
    arr = addArray(set, "PRIMARY", REAL32, dimensions = image%dims)
    optr=>real32Array2Data(arr)

    ! Copy image pixel data
    optr = image%image
!    optr = image%weight

    ! Add WCS keywords
    call setAttribute(arr, 'CTYPE1', 'RA---TAN',&
         & 'Coord. type: RA tangent plane projection')
    call setAttribute(arr, 'CRPIX1', image%xrpix, 'WCS reference pixel')
    call setAttribute(arr, 'CRVAL1', image%xrval,&
         & 'coord. at X axis ref. pixel')
    call setAttribute(arr, 'CDELT1', image%xdelt, 'WCS pixel size')
    call setAttribute(arr, 'CTYPE2', 'DEC--TAN',&
         & 'Coord. type: DEC tangent plane projection')
    call setAttribute(arr, 'CRPIX2', image%yrpix, 'WCS reference pixel')
    call setAttribute(arr, 'CRVAL2', image%yrval,&
         & 'coord. at Y axis ref. pixel')
    call setAttribute(arr, 'CDELT2', image%ydelt, 'WCS pixel size')

    ! Release dataset
    call release(arr)
    call release(set)
    
end subroutine writeImage

!********************************************************************
! Read an image as Real32. Assumes pointer points to correctly sized
! array of doubles.
!*******************************************************************
subroutine readImageAsReal32(arr, image)   ! Was real 64
    type (ArrayT), intent(in) :: arr
    real (kind=single), pointer, dimension(:,:) :: image

    ! Local variables
    real (kind=single), pointer :: f32array(:,:)
    real (kind=double), pointer :: f64array(:,:)
    integer (kind=int16), pointer :: i16array(:,:)
    integer (kind=int32), pointer :: i32array(:,:)
    integer :: dtype

    ! Perform type conversion on image data
    dtype=datatype(arr)
    select case (dtype)
    case (integer16)
       i16array=>int16Array2Data(arr)
       image = i16array
    case (integer32)
       i32array=>int32Array2Data(arr)
       image = i32array
    case (real32)
       f32array=>real32Array2Data(arr)
       image = f32array
    case (real64)
       f64array=>real64Array2Data(arr)
       image = f64array
    case default
       call error('badType', 'Image is of unsupported data type')
    end select
    
end subroutine readImageAsReal32



!**************************************************
! CheckFilters
! Checks that the filters in the files are all
! identical
!**************************************************
subroutine CheckFilters(nosw, oswFile)
    integer, intent(in) :: nosw
    integer, parameter :: nFilt = 7

    character(len=*), intent(in), dimension(1 : nosw) :: oswFile

    character(len=20), allocatable, dimension(:) :: filter
    character(len=20) :: oswFilter

    character(len = 100) :: string

    integer :: numFilters = 1
  
    integer :: i, j
       
    type (DataSetT) :: oswDataSet
    
    allocate(filter(1 : nosw))
 
    do i = 1 , nosw
        oswDataSet = DataSet(oswFile(i),read)
        oswFilter = trim(stringAttribute(oswDataSet, "FILTER"))

        if (i .eq. 1) then
            filter(1) = oswFilter
        else

            do j = 1, numFilters
        

                if (trim(oswFilter) .ne. trim(filter(j))) then
                    numFilters = numFilters + 1 
                    filter(numFilters) = oswFilter
                end if
      
            end do
        end if    
        call release(oswDataSet)
   
 
    end do
  
    if (numFilters .gt. 1) then
       string = "("
       do i = 1, numFilters - 1
           string = trim(string) // trim(filter(i)) // ", "
       end do
       
       string = trim(string) // trim(filter(numFilters)) // ")"

       call warning("You are combining images from different filters!", trim(string))
    end if


 
    deallocate(filter)

end subroutine CheckFilters


end module ommosaic_mod






